.drag-handle {
  color: alpha(@theme_fg_color, 0.4);
}

.drag-handle:backdrop {
  color: alpha(@theme_unfocused_fg_color, 0.4);
}

.split-row  picture:disabled,
.illustrated-row picture:disabled {
  /* FIXME: This is a workaround to handle disabled pictures
   * see: https: //gitlab.gnome.org/GNOME/libadwaita/-/merge_requests/814
   */
  opacity: 0.5;
}
