levelbar .strength-weak {
  background-color: #cc0000;
  border-color: #cc0000;
}

levelbar .strength-low {
  background-color: #f5ce00;
  border-color: #f5ce00;
}

levelbar .strength-medium,
levelbar .strength-good,
levelbar .strength-high {
  background-color: #73d216;
  border-color: #73d216;
}

/* This is used for user_avatar_edit_button */
.cutout-button {
  background-color: @window_bg_color;
  border-radius: 9999px;
  padding: 2px;
}
